#include <napi.h>
#include <stdio.h>
#include <iostream>
#include "filmlab/scene.hpp"
#include "filmlab/sceneview.hpp"
#include "filmlab/rawimage.hpp"
#include "filmlab/image.hpp"
#include "filmlab/imageloader.hpp"
#include "filmlab/renderer.hpp"
#include "filmlab/gl_helpers.hpp"
#include "filmlab/transforms.hpp"
#include <opencv2/opencv.hpp>

Napi::Uint8Array U8ArrayView(Napi::Env env, uint8_t* data, size_t elements) {
    Napi::ArrayBuffer buffer = Napi::ArrayBuffer::New(env, data, elements);
    return Napi::Uint8Array::New(env, elements, buffer, 0);
}

Napi::Uint16Array U16ArrayView(Napi::Env env, uint16_t* data, size_t elements) {
    Napi::ArrayBuffer buffer = Napi::ArrayBuffer::New(env, data, elements * sizeof(uint16_t));
    return Napi::Uint16Array::New(env, elements, buffer, 0);
}

Napi::Float32Array FloatArrayView(Napi::Env env, float* data, size_t elements) {
    Napi::ArrayBuffer buffer = Napi::ArrayBuffer::New(env, data, elements * sizeof(float));
    return Napi::Float32Array::New(env, elements, buffer, 0);
}

Napi::Float64Array DoubleArrayView(Napi::Env env, double* data, size_t elements) {
    Napi::ArrayBuffer buffer = Napi::ArrayBuffer::New(env, data, elements * sizeof(double));
    return Napi::Float64Array::New(env, elements, buffer, 0);
}

Napi::Object ImageFormatWrapper(Napi::Env env, filmlab::ImageFormat& formatData) {
  Napi::Object imageFormat = Napi::Object::New(env);
  imageFormat.Set("orientation", static_cast<int>(formatData.orientation));
  imageFormat.Set("colorSpace", static_cast<int>(formatData.colorSpace));
  imageFormat.Set("pixelOrder", static_cast<int>(formatData.pixelOrder));
  imageFormat.Set("gammaCurve", static_cast<int>(formatData.gammaCurve));
  //imageFormat.Set("illuminantXY", DoubleArrayView(env, formatData.illuminantXY, 2));
  //imageFormat.Set("transformMatrixXYZ", DoubleArrayView(env, formatData.transformMatrixXYZ[0], 9));
  return imageFormat;
}

Napi::Object MatWrapper(Napi::Env env, cv::Mat& mat) {
    Napi::Object wrappedMat = Napi::Object::New(env);
    wrappedMat.Set("rows", mat.rows);
    wrappedMat.Set("cols", mat.cols);
    wrappedMat.Set("type", mat.type());
    wrappedMat.Set("channels", mat.channels());
    wrappedMat.Set("depth", mat.depth());
    if (mat.type() == CV_16U || mat.type() == CV_16F) {
      wrappedMat.Set("data", U16ArrayView(env, mat.ptr<uint16>(), mat.total() * mat.channels()));
    } else if (mat.type() == CV_32F) {
      wrappedMat.Set("data", FloatArrayView(env, mat.ptr<float>(), mat.total() * mat.channels()));
    } else if (mat.type() == CV_64F) {
      wrappedMat.Set("data", DoubleArrayView(env, mat.ptr<double>(), mat.total() * mat.channels()));
    } else {
      wrappedMat.Set("data", U8ArrayView(env, mat.ptr<uchar>(), mat.total() * mat.elemSize()));
    }
   
    return wrappedMat;
}

Napi::Object ImageWrapper(Napi::Env env, filmlab::Image& image) {
   Napi::Object wrappedImage = Napi::Object::New(env);
   wrappedImage.Set("format", ImageFormatWrapper(env, image.format));
   wrappedImage.Set("mat", MatWrapper(env, image.mat));
   return wrappedImage;
}

filmlab::Scene* scene = nullptr;
filmlab::SceneView* sceneView;
std::string currentImagePath;
Napi::ThreadSafeFunction jsSceneChangeHandler;
//Napi::ObjectReference jsSceneChangeHandlerThis;
std::function<void()> sceneChangeHandler = []() {
  jsSceneChangeHandler.BlockingCall();
};
float viewportTransform[3][3];
cv::Rect2d viewportCrop;

class JSBridgeRenderer: public filmlab::ColorRenderer {
    public:
    float viewportTransform[3][3];

   Napi::Object getRendererData(Napi::Env env) {
      //printf("💚NODE: calculating matrices");
      Napi::Object jsData = Napi::Object::New(env);
      //calculateMatrices();
      
      Napi::Object lut = Napi::Object::New(env);      
      lut.Set("empty", workingColorSpaceToXYZLUT.data.empty());
      lut.Set("bins", workingColorSpaceToXYZLUT.bins);
      cv::Mat mat =  workingColorSpaceToXYZLUT.data;
      lut.Set("data", FloatArrayView(
         env,
         workingColorSpaceToXYZLUT.data.ptr<float>(),
         pow(workingColorSpaceToXYZLUT.bins, 3) * 3));
      lut.Set("inputMin", DoubleArrayView(env, workingColorSpaceToXYZLUT.inputMin, 3));
      lut.Set("inputMax", DoubleArrayView(env, workingColorSpaceToXYZLUT.inputMax, 3));

      jsData.Set("workingColorSpaceToXYZLUT", lut);

      jsData.Set("imageFormat", ImageFormatWrapper(env, formatData));

      //jsData.Set("inputTargetWhitePointXY", DoubleArrayView(env, inputTargetWhitePointXY, 2));
      jsData.Set("outputBrightness", outputBrightness);
      jsData.Set("outputContrast", outputContrast);
      jsData.Set("xyzToOutputColorTransformMatrix", DoubleArrayView(env, xyzToOutputColorTransformMatrix[0], 9));
      jsData.Set("workingColorSpaceLUTGammaCurve", static_cast<int>(workingColorSpaceLUTGammaCurve));

      jsData.Set("inputImageColorSpaceToWorkingColorSpaceMatrix", DoubleArrayView(env, inputImageColorSpaceToWorkingColorSpaceMatrix[0], 9));
            
      jsData.Set("imageFormatChanged", imageFormatChanged);
      jsData.Set("workingColorSpaceToXYZLUTChanged", workingColorSpaceToXYZLUTChanged);
      jsData.Set("adjustmentsChanged", adjustmentsChanged);

      // Calculate 3D transform from source image to viewport
      cv::Rect2d viewCrop;
      cv::Mat transform = sceneView->getSceneImageToViewportTransform(true, &viewCrop);
      cv::Size rendererSize = sceneView->getViewportSize();
      transform = filmlab::makeScaleMatrix(1.0 / rendererSize.width, 1.0 / rendererSize.height) *
          transform * filmlab::makeScaleMatrix(scene->inputImage.mat.cols, scene->inputImage.mat.rows);
      transform = filmlab::makeVerticalFlipMatrix(1.0) * transform; // Y-flipped for GL
      // Invert transform so it can be used as a lookup from output instead of a transform of input
      transform = filmlab::invertTransformMatrix(transform);
      transform.convertTo(transform, CV_32F);
      memcpy(viewportTransform, transform.ptr<float>(), 9 * sizeof(float));
      jsData.Set("viewTransform", FloatArrayView(env, viewportTransform[0], 9));
      jsData.Set("viewCropX", viewCrop.x);
      jsData.Set("viewCropY", viewCrop.y);
      jsData.Set("viewCropWidth", viewCrop.width);
      jsData.Set("viewCropHeight", viewCrop.height);

      return jsData;

   }
};

JSBridgeRenderer renderer;

void createScene() {
    if (scene != nullptr) {
        delete scene;
        scene = nullptr;
    }
    if (sceneView != nullptr) {
        delete sceneView;
        sceneView = nullptr;
    }
    scene = new filmlab::Scene();
    scene->useAsyncThreads = true;
    scene->useAutoMask = true;
    scene->autoSettingsCalculationInterval = 33;
    sceneView = new filmlab::SceneView(scene);
}

void loadImage(const Napi::CallbackInfo& info) {
  std::string path = info[0].As<Napi::String>();
  printf("%s\n", path.c_str());
  filmlab::ImageLoader loader;

  filmlab::Image image;
  if (loader.loadImageFromPath(path, image)) {
    ////image.mat.setTo(image.mat.depth() == CV_8U ? cv::Scalar(255, 255, 0) : cv::Scalar(65535, 65535, 0));
    scene->setImage(image);
    scene->applyChanges();
    currentImagePath = path;
  }
}

Napi::Object getTextureImage(const Napi::CallbackInfo& info) {
  Napi::Env env = info.Env();
  return ImageWrapper(env, scene->inputImage);
}

void saveJPEG(const Napi::CallbackInfo& info) {
    if (currentImagePath.length() == 0) return;
    cv::Mat out = scene->makeProcessedImage();
    out.convertTo(out, CV_8U, 1.0/256.0);
    std::vector<int> params = {cv::IMWRITE_JPEG_QUALITY, 95};
    cv::imwrite(currentImagePath + ".jpg", out, params);
}

void setViewportSize(const Napi::CallbackInfo& info) {
    int width = info[0].As<Napi::Number>().Int32Value();
    int height = info[1].As<Napi::Number>().Int32Value();
    if (sceneView != nullptr) {
        sceneView->setViewportSize(cv::Size(width, height));
        sceneView->setViewportSafeArea(cv::Rect2d(0, 0, width, height));
        sceneView->setViewportOrientation(filmlab::Orientation::AS_IS);
    }
    printf("Set view size %d, %d\n", width, height);
}

void setSceneSetting(const Napi::CallbackInfo& info) {
    int setting = info[0].As<Napi::Number>().Int32Value();
    double value = info[1].As<Napi::Number>().DoubleValue();
    scene->setSetting(filmlab::ProcessSetting(setting), value);
    scene->applyChanges();
}

napi_value getSceneSetting(const Napi::CallbackInfo& info) {
    Napi::Env env = info.Env();
    int setting = info[0].As<Napi::Number>().Int32Value();
    return Napi::Number::New(env, scene->getSetting(filmlab::ProcessSetting(setting)));
}

napi_value getImageWidth(const Napi::CallbackInfo& info) {
    Napi::Env env = info.Env();
    return Napi::Number::New(env, scene->inputImage.mat.cols);
}

napi_value getImageHeight(const Napi::CallbackInfo& info) {
    Napi::Env env = info.Env();
    return Napi::Number::New(env, scene->inputImage.mat.rows);
}

void zoomIn(const Napi::CallbackInfo& info) {
  sceneView->adjustViewportZoom(1.5);
}

void zoomOut(const Napi::CallbackInfo& info) {
  sceneView->adjustViewportZoom(1.0/1.5);
}

void setZoom(const Napi::CallbackInfo& info) {
  double zoom = info[0].As<Napi::Number>().DoubleValue();
  sceneView->adjustViewportZoom(zoom);
}

void adjustPan(const Napi::CallbackInfo& info) {
  assert(info.Length() == 2);
  double x = info[0].As<Napi::Number>().DoubleValue();
  double y = info[1].As<Napi::Number>().DoubleValue();
  sceneView->adjustViewportPan(cv::Point2d(x, y));
}

void flipHorizontally(const Napi::CallbackInfo& info) {
  scene->settings.flipHorizontally();
  scene->applyChanges();
}

void rotateCounterClockwise(const Napi::CallbackInfo& info) {
  scene->settings.rotateCounterClockwise();
  scene->applyChanges();
}

napi_value hasColorControl(const Napi::CallbackInfo& info) {
    if (scene == nullptr || scene->process == nullptr) {
        return Napi::Boolean::New(info.Env(), false);
    }
    return Napi::Boolean::New(info.Env(),
        scene->process->hasColorTempControl() &&
        scene->process->hasColorTintControl());
}

Napi::Object getRendererData(const Napi::CallbackInfo& info) {
    if (scene != nullptr && scene->process != nullptr) {
        scene->process->updateRenderer(scene->inputImage.format, &renderer);
    }
    return renderer.getRendererData(info.Env());
}

void setSceneChangeHandler(const Napi::CallbackInfo& info) {
    Napi::Env env = info.Env();
    Napi::Function callback = info[0].As<Napi::Function>();
    jsSceneChangeHandler = Napi::ThreadSafeFunction::New(
      env, callback, "change handler thread", 0, 1, [](Napi::Env) {});
    scene->changeDelegate = &sceneChangeHandler;
    sceneView->changeDelegate = &sceneChangeHandler;
}

Napi::Object Init(Napi::Env env, Napi::Object exports) {
  printf("FilmLab native layer initializing.\n");

  createScene();

  exports.Set("loadImage",
              Napi::Function::New(env, loadImage));

  exports.Set("saveJPEG",
              Napi::Function::New(env, saveJPEG));

  exports.Set("setViewportSize",
              Napi::Function::New(env, setViewportSize));

  exports.Set("setSceneSetting",
              Napi::Function::New(env, setSceneSetting));

  exports.Set("getSceneSetting",
              Napi::Function::New(env, getSceneSetting));

  exports.Set("getImageWidth",
              Napi::Function::New(env, getImageWidth));

  exports.Set("getImageHeight",
              Napi::Function::New(env, getImageHeight));

  exports.Set("setZoom",
              Napi::Function::New(env, setZoom));

  exports.Set("zoomIn",
              Napi::Function::New(env, zoomIn));

  exports.Set("zoomOut",
              Napi::Function::New(env, zoomOut));  

  exports.Set("adjustPan",
              Napi::Function::New(env, adjustPan));

  exports.Set("flipHorizontally",
              Napi::Function::New(env, flipHorizontally));

  exports.Set("rotateCounterClockwise",
              Napi::Function::New(env, rotateCounterClockwise));  

  exports.Set("hasColorControl",
              Napi::Function::New(env, hasColorControl));    

  exports.Set("getRendererData", Napi::Function::New(env, getRendererData));

  exports.Set("getTextureImage", Napi::Function::New(env, getTextureImage));  

  exports.Set("setSceneChangeHandler", Napi::Function::New(env, setSceneChangeHandler));

  return exports;
}

NODE_API_MODULE(NODE_GYP_MODULE_NAME, Init)
